"""
Wind farm optimization objectives.
"""

from .farm_vars import FarmVarObjective as FarmVarObjective
from .farm_vars import MaxFarmPower as MaxFarmPower
from .farm_vars import MinimalMaxTI as MinimalMaxTI

from .max_n_turbines import MaxNTurbines as MaxNTurbines
